#include <gtk/gtk.h>
#include "specimen.h"
#include "patch.h"
#include "gui.h"
#include "lfo.h"

/* parameter codes */
enum
{
     DELAY,
     ATTACK,
     HOLD,
     DECAY,
     SUSTAIN,
     RELEASE,
     AMOUNT,
     LFO_AMOUNT,
     VEL_AMOUNT,
};

/* different possible beat lengths */
enum
{
     EIGHTH,
     QUARTER,
     HALF,
     WHOLE,
     DOUBLE,
};

/* track which parameter's envelope we're setting */
static PatchParamType cur_param = PATCH_PARAM_VOLUME;

/* widgets */
static GtkWidget* window;
static GtkWidget* e_spin;	/* dElay */
static GtkWidget* a_spin;
static GtkWidget* h_spin;
static GtkWidget* d_spin;
static GtkWidget* s_spin;
static GtkWidget* r_spin;
static GtkWidget* amt_spin;
static GtkWidget* amt_label;
static GtkWidget* lfo_delay_spin;
static GtkWidget* lfo_attack_spin;
static GtkWidget* lfo_amt_spin;
static GtkWidget* lfo_freq_spin;
static GtkWidget* lfo_shape_opt;
static GtkWidget* lfo_sync_opt;
static GtkWidget* lfo_sync_check;
static GtkWidget* lfo_global_check;
static GtkWidget* lfo_positive_check;
static GtkWidget* vel_amt_spin;

/* adjustments */
static GtkAdjustment* e_adj;
static GtkAdjustment* a_adj;
static GtkAdjustment* h_adj;
static GtkAdjustment* d_adj;
static GtkAdjustment* s_adj;
static GtkAdjustment* r_adj;
static GtkAdjustment* amt_adj;
static GtkAdjustment* pitch_amt_adj;
static GtkAdjustment* lfo_delay_adj;
static GtkAdjustment* lfo_attack_adj;
static GtkAdjustment* lfo_amt_adj;
static GtkAdjustment* lfo_pitch_amt_adj;
static GtkAdjustment* lfo_freq_adj;
static GtkAdjustment* vel_amt_adj;

static void update_lfo (int cp)
{
     float freq;
     float sync;
     float attack;
     float delay;
     gboolean sync_enable;
     gboolean global;
     gboolean positive;
     LFOShape shape;

     if (patch_get_lfo_attack (cp, cur_param, &attack) < 0)
     {
	  debug ("failed to get lfo attack\n");
     }

     if (patch_get_lfo_delay (cp, cur_param, &delay) < 0)
     {
	  debug ("failed to get lfo delay\n");
     }

     if (patch_get_lfo_freq (cp, cur_param, &freq) < 0)
     {
	  debug ("failed to get frequency\n");
     }

     if (patch_get_lfo_shape (cp, cur_param, &shape) < 0)
     {
	  debug ("failed to get shape\n");
	  return;
     }

     if (patch_get_lfo_sync (cp, cur_param, &sync_enable) < 0)
     {
	  debug ("failed to get sync\n");
	  return;
     }

     if (patch_get_lfo_beats (cp, cur_param, &sync) < 0)
     {
	  debug ("failed to get sync beats\n");
	  return;
     }

     if (patch_get_lfo_global (cp, cur_param, &global) < 0)
     {
	  debug ("failed to get globalness\n");
	  return;
     }
     
     if (patch_get_lfo_positive (cp, cur_param, &positive) < 0)
     {
	  debug ("failed to get positiveness\n");
	  return;
     }
     
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (lfo_freq_spin),
				(gdouble) freq);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (lfo_attack_spin),
				(gdouble) attack);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (lfo_delay_spin),
				(gdouble) delay);

     switch (shape)
     {
     case LFO_SHAPE_TRIANGLE:
	  gtk_option_menu_set_history (GTK_OPTION_MENU (lfo_shape_opt), 0);
	  break;
     case LFO_SHAPE_SINE:
	  gtk_option_menu_set_history (GTK_OPTION_MENU (lfo_shape_opt), 1);
	  break;
     case LFO_SHAPE_SAW:
	  gtk_option_menu_set_history (GTK_OPTION_MENU (lfo_shape_opt), 2);
	  break;
     case LFO_SHAPE_SQUARE:
	  gtk_option_menu_set_history (GTK_OPTION_MENU (lfo_shape_opt), 3);
	  break;
     }


     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (lfo_sync_check), sync_enable);

     /* we need to make sure that cb_sync gets called and updates the
      * sensitivity of lfo_freq_spin and lfo_sync_opt because this may
      * be our first time being called (so cb_sync hasn't been called
      * at all yet) and the patch might not be synced (and since
      * lfo_sync_check is unchecked by default, setting it to
      * unchecked wouldn't toggle it, leaving lfo_sync_opt sensitive
      * when it shouldn't be) */
     g_signal_emit_by_name (G_OBJECT (lfo_sync_check), "toggled");

     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (lfo_global_check), global);
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (lfo_positive_check), positive);
     
     if (sync == 0.5)
     {
	  gtk_option_menu_set_history (GTK_OPTION_MENU (lfo_sync_opt), 0);

     }
     else if (sync == 1.0)
     {
	  gtk_option_menu_set_history (GTK_OPTION_MENU (lfo_sync_opt), 1);

     }
     else if (sync == 2.0)
     {
	  gtk_option_menu_set_history (GTK_OPTION_MENU (lfo_sync_opt), 2);

     }
     else if (sync == 4.0)
     {
	  gtk_option_menu_set_history (GTK_OPTION_MENU (lfo_sync_opt), 3);

     }
     else
     {
	  gtk_option_menu_set_history (GTK_OPTION_MENU (lfo_sync_opt), 4);
     }


}

static int update_interface ( )
{
     int cp;
     float e, a, h, d, s, r;
     float amt, lfo_amt, vel_amt;

     if ((cp = get_current_patch ( )) < 0)
	  return -1;

     patch_get_env_delay   (cp, cur_param, &e);
     patch_get_env_attack  (cp, cur_param, &a);
     patch_get_env_hold    (cp, cur_param, &h);
     patch_get_env_decay   (cp, cur_param, &d);
     patch_get_env_sustain (cp, cur_param, &s);
     patch_get_env_release (cp, cur_param, &r);
     patch_get_lfo_amount  (cp, cur_param, &lfo_amt);
     patch_get_vel_amount  (cp, cur_param, &vel_amt);

     gtk_spin_button_set_value (GTK_SPIN_BUTTON (e_spin), e);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (a_spin), a);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (h_spin), h);     
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (d_spin), d);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (s_spin), s);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (r_spin), r);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (vel_amt_spin), vel_amt);

     if (cur_param == PATCH_PARAM_PITCH)
	  lfo_amt *= PATCH_MAX_PITCH_STEPS;

     gtk_spin_button_set_value (GTK_SPIN_BUTTON (lfo_amt_spin), lfo_amt);
     
     /* volume has no envelope amount setting */
     if (cur_param == PATCH_PARAM_VOLUME)
     {
	  gtk_widget_hide (amt_spin);
	  gtk_widget_hide (amt_label);
     }
     else
     {
	  patch_get_env_amount (cp, cur_param, &amt);
	  if (cur_param == PATCH_PARAM_PITCH)
	       amt *= PATCH_MAX_PITCH_STEPS;

	  gtk_spin_button_set_value (GTK_SPIN_BUTTON (amt_spin), amt * PATCH_MAX_PITCH_STEPS);
	  gtk_widget_show (amt_spin);
	  gtk_widget_show (amt_label);
	  gtk_widget_queue_draw (amt_label);	/* TODO: figure out why this doesn't always work */
     }

     /* update the lfo specific parts of our interface */
     update_lfo (cp);
     return 0;
}

static void cb_close ( )
{
     gtk_widget_hide (window);
}

static void cb_set (GtkWidget*  widget, int param)
{
     int cp, err;
     double val;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     switch (param)
     {
     case DELAY:
	  val =  gtk_spin_button_get_value (GTK_SPIN_BUTTON (e_spin));

	  if ((err = patch_set_env_delay (cp, cur_param, val)) < 0)
	       errmsg ("Failed to set delay for patch %d (%s)\n", cp,
		       patch_strerror (err));
	  break;
     case ATTACK:
	  val =  gtk_spin_button_get_value (GTK_SPIN_BUTTON (a_spin));

	  if ((err = patch_set_env_attack (cp, cur_param, val)) < 0)
	       errmsg ("Failed to set attack for patch %d (%s)\n", cp,
		       patch_strerror (err));
	  break;
     case HOLD:
	  val =  gtk_spin_button_get_value (GTK_SPIN_BUTTON (h_spin));

	  if ((err = patch_set_env_hold (cp, cur_param, val)) < 0)
	       errmsg ("Failed to set hold for patch %d (%s)\n", cp,
		       patch_strerror (err));
	  break;
     case DECAY:
	  val =  gtk_spin_button_get_value (GTK_SPIN_BUTTON (d_spin));

	  if ((err = patch_set_env_decay (cp, cur_param, val)) < 0)
	       errmsg ("Failed to set decay for patch %d (%s)\n", cp,
		       patch_strerror (err));
	  break;
     case SUSTAIN:
	  val =  gtk_spin_button_get_value (GTK_SPIN_BUTTON (s_spin));

	  if ((err = patch_set_env_sustain (cp, cur_param, val)) < 0)
	       errmsg ("Failed to set sustain for patch %d (%s)\n", cp,
		       patch_strerror (err));
	  break;
     case RELEASE:
	  val =  gtk_spin_button_get_value (GTK_SPIN_BUTTON (r_spin));

	  if ((err = patch_set_env_release (cp, cur_param, val)) < 0)
	       errmsg ("Failed to set release for patch %d (%s)\n", cp,
		       patch_strerror (err));
	  break;
     case AMOUNT:
	  val =  gtk_spin_button_get_value (GTK_SPIN_BUTTON (amt_spin));

	  if (cur_param == PATCH_PARAM_PITCH)
	       val /= PATCH_MAX_PITCH_STEPS;
	  if ((err = patch_set_env_amount (cp, cur_param, val)) < 0)
	       errmsg ("Failed to set release for patch %d (%s)\n", cp,
		       patch_strerror (err));
	  break;
     case LFO_AMOUNT:
	  val =  gtk_spin_button_get_value (GTK_SPIN_BUTTON (lfo_amt_spin));

	  if (cur_param == PATCH_PARAM_PITCH)
	       val /= PATCH_MAX_PITCH_STEPS;
	  if ((err = patch_set_lfo_amount (cp, cur_param, val)) < 0)
	       errmsg ("Failed to set lfo amount for patch %d (%s)\n", cp,
		       patch_strerror (err));
	  break;
     case VEL_AMOUNT:
	  val =  gtk_spin_button_get_value (GTK_SPIN_BUTTON (vel_amt_spin));

	  if ((err = patch_set_vel_amount (cp, cur_param, val)) < 0)
	       errmsg ("Failed to set velocity amount for patch %d (%s)\n", cp,
		       patch_strerror (err));
	  break;
     }

     return;
}

static void cb_freq (GtkWidget* widget, gpointer data)
{
     int cp;
     float freq;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     freq = gtk_spin_button_get_value (GTK_SPIN_BUTTON (lfo_freq_spin));

     if (patch_set_lfo_freq (cp, cur_param, freq) < 0)
	  debug ("failed to set freq\n");
}

static void cb_shape (GtkWidget* widget, LFOShape shape)
{
     int cp;
     
     if ((cp = get_current_patch ( )) < 0)
	  return;

     if (patch_set_lfo_shape (cp, cur_param, shape) < 0)
	  debug ("failed to set shape\n");
}

static void cb_sync (GtkWidget* widget, gpointer data)
{
     int cp;
     
     if ((cp = get_current_patch ( )) < 0)
	  return;

     if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
     {
	  gtk_widget_set_sensitive (lfo_freq_spin, FALSE);
	  gtk_widget_set_sensitive (lfo_sync_opt, TRUE);

	  if (patch_set_lfo_sync (cp, cur_param, TRUE) < 0)
	       debug ("failed to set sync\n");
     }
     else
     {
	  gtk_widget_set_sensitive (lfo_freq_spin, TRUE);
	  gtk_widget_set_sensitive (lfo_sync_opt, FALSE);

	  if (patch_set_lfo_sync (cp, cur_param, FALSE) < 0)
	       debug ("failed to set sync\n");
     }
}

static void cb_sync_beats (GtkWidget* widget, int beats)
{
     int cp;
     float len;
     
     if ((cp = get_current_patch ( )) < 0)
	  return;

     switch (beats)
     {
     case EIGHTH:
	  len = 0.5;
	  break;
     case QUARTER:
	  len = 1.0;
	  break;
     case HALF:
	  len = 2.0;
	  break;
     case WHOLE:
	  len = 4.0;
	  break;
     case DOUBLE:
	  len = 8.0;
	  break;
     default:
	  debug ("invalid length specified\n");
	  return;
     }

     if (patch_set_lfo_beats (cp, cur_param, len) < 0)
	  debug ("failed to set sync beats\n");
}

static void cb_lfo_attack (GtkWidget* widget, gpointer data)
{
     int cp;
     float val;
     
     if ((cp = get_current_patch ( )) < 0)
	  return;

     val = gtk_spin_button_get_value (GTK_SPIN_BUTTON (lfo_attack_spin));
     if (patch_set_lfo_attack (cp, cur_param, val) < 0)
     {
	  debug ("failed to set lfo attack\n");
     }
}

static void cb_lfo_delay (GtkWidget* widget, gpointer data)
{
     int cp;
     float val;
     
     if ((cp = get_current_patch ( )) < 0)
	  return;

     val = gtk_spin_button_get_value (GTK_SPIN_BUTTON (lfo_delay_spin));
     if (patch_set_lfo_delay (cp, cur_param, val) < 0)
     {
	  debug ("failed to set lfo delay\n");
     }
}

static void cb_global (GtkWidget* widget, gpointer data)
{
     int cp;
     
     if ((cp = get_current_patch ( )) < 0)
	  return;

     if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (lfo_global_check)))
     {
	  patch_set_lfo_global (cp, cur_param, TRUE);
     }
     else
     {
	  patch_set_lfo_global (cp, cur_param, FALSE);
     }
}

static void cb_positive (GtkWidget* widget, gpointer data)
{
     int cp;
     
     if ((cp = get_current_patch ( )) < 0)
	  return;

     if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (lfo_positive_check)))
     {
	  patch_set_lfo_positive (cp, cur_param, TRUE);
     }
     else
     {
	  patch_set_lfo_positive (cp, cur_param, FALSE);
     }
}

static void cb_param (GtkWidget* widget, PatchParamType param)
{
     cur_param = param;
     if (param == PATCH_PARAM_PITCH)
     {
	  gtk_spin_button_set_adjustment (GTK_SPIN_BUTTON (lfo_amt_spin), lfo_pitch_amt_adj);
	  gtk_spin_button_set_adjustment (GTK_SPIN_BUTTON (amt_spin), pitch_amt_adj);
     }
     else
     {
	  gtk_spin_button_set_adjustment (GTK_SPIN_BUTTON (lfo_amt_spin), lfo_amt_adj);
	  gtk_spin_button_set_adjustment (GTK_SPIN_BUTTON (amt_spin), amt_adj);
     }
     update_interface ( );
}

void param_settings_show ( )
{
     if (update_interface ( ) == 0)
	  gtk_widget_show (window);
}

void param_settings_init (GtkWidget*  parent)
{
     GtkWidget* vbox;
     GtkWidget* hbox;
     GtkWidget* button;
     GtkWidget* label;
     GtkWidget* sep;
     GtkWidget* menu;
     GtkWidget* opt;
     GtkWidget* item;

     debug ("Initializing parameter settings window\n");

     /* main window */
     window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
     gtk_window_set_title (GTK_WINDOW (window), "Params");
     gtk_window_set_resizable (GTK_WINDOW (window), FALSE);
     gtk_window_set_transient_for (GTK_WINDOW (window),
				   GTK_WINDOW (parent));
     gtk_window_set_modal (GTK_WINDOW (window), TRUE);
     g_signal_connect (GTK_WINDOW (window), "delete-event",
		       G_CALLBACK (cb_close), NULL);

     /* vbox */
     vbox = gtk_vbox_new (FALSE, SPACING);
     gtk_container_add (GTK_CONTAINER (window), vbox);
     gtk_container_set_border_width (GTK_CONTAINER (window), SPACING);
     gtk_widget_show (vbox);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /* parameter button menu */
     menu = gtk_menu_new ( );

     /* volume */
     item = gtk_menu_item_new_with_label ("Volume");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (cb_param),
		       (gpointer) PATCH_PARAM_VOLUME);
     gtk_widget_show (item);

     /* panning */
     item = gtk_menu_item_new_with_label ("Panning");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (cb_param),
		       (gpointer) PATCH_PARAM_PANNING);
     gtk_widget_show (item);

     /* cutoff freq */
     item = gtk_menu_item_new_with_label ("Cutoff");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (cb_param),
		       (gpointer) PATCH_PARAM_CUTOFF);
     gtk_widget_show (item);

     /* resonance */
     item = gtk_menu_item_new_with_label ("Resonance");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (cb_param),
		       (gpointer) PATCH_PARAM_RESONANCE);
     gtk_widget_show (item);

     /* pitch */
     item = gtk_menu_item_new_with_label ("Pitch");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (cb_param),
		       (gpointer) PATCH_PARAM_PITCH);
     gtk_widget_show (item);

     /* parameter optionmenu */
     opt = gtk_option_menu_new ( );
     gtk_box_pack_start (GTK_BOX (hbox), opt, FALSE, FALSE, 0);
     gtk_option_menu_set_menu (GTK_OPTION_MENU (opt), menu);
     gtk_widget_show (opt);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /***************** ADSR PARAMS ********************/
     label = gtk_label_new (NULL);
     gtk_label_set_markup (GTK_LABEL (label), "<b>ADSR</b>");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /*********/
     /* delay */
     /*********/
     label = gtk_label_new ("Delay:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* adjustment */
     e_adj =  (GtkAdjustment *) gtk_adjustment_new (0.0, 0.0, 3.0, 0.01, 0.1,
						    0.0);
     g_signal_connect (G_OBJECT (e_adj), "value-changed",
		       G_CALLBACK (cb_set), (gpointer) DELAY);

     /* spin */
     e_spin = gtk_spin_button_new (e_adj, 0, 2);
     gtk_box_pack_end (GTK_BOX (hbox), e_spin, FALSE, FALSE, 0);
     gtk_widget_show (e_spin);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /**********/
     /* attack */
     /**********/
     label = gtk_label_new ("Attack:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* adjustment */
     a_adj =  (GtkAdjustment *) gtk_adjustment_new (0.0, 0.0, 3.0, 0.01, 0.1,
						    0.0);
     g_signal_connect (G_OBJECT (a_adj), "value-changed",
		       G_CALLBACK (cb_set), (gpointer) ATTACK);

     /* spin */
     a_spin = gtk_spin_button_new (a_adj, 0, 2);
     gtk_box_pack_end (GTK_BOX (hbox), a_spin, FALSE, FALSE, 0);
     gtk_widget_show (a_spin);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /********/
     /* hold */
     /********/
     label = gtk_label_new ("Hold:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* adjustment */
     h_adj =  (GtkAdjustment *) gtk_adjustment_new (0.0, 0.0, 3.0, 0.01, 0.1,
						    0.0);
     g_signal_connect (G_OBJECT (h_adj), "value-changed",
		       G_CALLBACK (cb_set), (gpointer) HOLD);

     /* spin */
     h_spin = gtk_spin_button_new (h_adj, 0, 2);
     gtk_box_pack_end (GTK_BOX (hbox), h_spin, FALSE, FALSE, 0);
     gtk_widget_show (h_spin);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /*********/
     /* decay */
     /*********/
     label = gtk_label_new ("Decay:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* adjustment */
     d_adj = (GtkAdjustment *) gtk_adjustment_new (0.0, 0.0, 3.0, 0.01, 0.1,
						   0.0);
     g_signal_connect (G_OBJECT (d_adj), "value-changed",
		       G_CALLBACK (cb_set), (gpointer) DECAY);

     /* spin */
     d_spin = gtk_spin_button_new (d_adj, 0, 2);
     gtk_box_pack_end (GTK_BOX (hbox), d_spin, FALSE, FALSE, 0);
     gtk_widget_show (d_spin);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /***********/
     /* sustain */
     /***********/
     label = gtk_label_new ("Sustain:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* adjustment */
     s_adj = (GtkAdjustment *) gtk_adjustment_new (0.0, 0.0, 1.0, 0.01, 0.1,
						   0.0);
     g_signal_connect (G_OBJECT (s_adj), "value-changed",
		       G_CALLBACK (cb_set), (gpointer) SUSTAIN);

     /* spin */
     s_spin = gtk_spin_button_new (s_adj, 0, 2);
     gtk_box_pack_end (GTK_BOX (hbox), s_spin, FALSE, FALSE, 0);
     gtk_widget_show (s_spin);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /***********/
     /* release */
     /***********/
     label = gtk_label_new ("Release:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* adjustment */
     r_adj = (GtkAdjustment *) gtk_adjustment_new (0.0, 0.0, 3.0, 0.01, 0.1,
						   0.0);
     g_signal_connect (G_OBJECT (r_adj), "value-changed",
		       G_CALLBACK (cb_set), (gpointer) RELEASE);

     /* spin */
     r_spin = gtk_spin_button_new (r_adj, 0, 2);
     gtk_box_pack_end (GTK_BOX (hbox), r_spin, FALSE, FALSE, 0);
     gtk_widget_show (r_spin);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /***********/
     /* amount */
     /***********/
     amt_label = gtk_label_new ("Amount:");
     gtk_box_pack_start (GTK_BOX (hbox), amt_label, FALSE, FALSE, 0);
     gtk_widget_show (amt_label);

     /* adjustment */
     amt_adj = (GtkAdjustment *) gtk_adjustment_new (0.0, -1.0, 1.0, 0.01, 0.1, 0.0);
     pitch_amt_adj = (GtkAdjustment *) gtk_adjustment_new (0.0, -PATCH_MAX_PITCH_STEPS, PATCH_MAX_PITCH_STEPS,
						     0.01, 0.1, 0.0);
     g_signal_connect (G_OBJECT (amt_adj), "value-changed",
		       G_CALLBACK (cb_set), (gpointer) AMOUNT);
     g_signal_connect (G_OBJECT (pitch_amt_adj), "value-changed",
		       G_CALLBACK (cb_set), (gpointer) AMOUNT);

     /* spin */
     amt_spin = gtk_spin_button_new (amt_adj, 0, 2);
     gtk_box_pack_end (GTK_BOX (hbox), amt_spin, FALSE, FALSE, 0);
     gtk_widget_show (amt_spin);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /***************** LFO PARAMS ********************/
     label = gtk_label_new (NULL);
     gtk_label_set_markup (GTK_LABEL (label), "<b>LFO</b>");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /**********/
     /* Delay */
     /**********/
     label = gtk_label_new ("Delay:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* adjustment */
     lfo_delay_adj = (GtkAdjustment *) gtk_adjustment_new (1.0, 0.0, 3.0, 0.01, 0.1, 1.0);

     /* spin */
     lfo_delay_spin = gtk_spin_button_new (lfo_delay_adj, 0.0, 2);
     gtk_box_pack_end (GTK_BOX (hbox), lfo_delay_spin, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (lfo_delay_adj), "value-changed",
		       G_CALLBACK (cb_lfo_delay), NULL);
     gtk_widget_show (lfo_delay_spin);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /**********/
     /* Attack */
     /**********/
     label = gtk_label_new ("Attack:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* adjustment */
     lfo_attack_adj = (GtkAdjustment *) gtk_adjustment_new (1.0, 0.0, 3.0, 0.01, 0.1, 1.0);

     /* spin */
     lfo_attack_spin = gtk_spin_button_new (lfo_attack_adj, 0.0, 2);
     gtk_box_pack_end (GTK_BOX (hbox), lfo_attack_spin, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (lfo_attack_adj), "value-changed",
		       G_CALLBACK (cb_lfo_attack), NULL);
     gtk_widget_show (lfo_attack_spin);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /********/
     /* Freq */
     /********/
     label = gtk_label_new ("Freq:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* adjustment */
     lfo_freq_adj = (GtkAdjustment *) gtk_adjustment_new (1.0, 0.0, 20.0, 0.01, 0.1, 1.0);

     /* spin */
     lfo_freq_spin = gtk_spin_button_new (lfo_freq_adj, 0.0, 2);
     gtk_box_pack_end (GTK_BOX (hbox), lfo_freq_spin, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (lfo_freq_adj), "value-changed",
		       G_CALLBACK (cb_freq), NULL);
     gtk_widget_show (lfo_freq_spin);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /*********/
     /* Sync  */
     /*********/
     label = gtk_label_new ("Sync:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* sync check box */
     lfo_sync_check = gtk_check_button_new ( );
     gtk_box_pack_start (GTK_BOX (hbox), lfo_sync_check, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (lfo_sync_check), "toggled",
		       G_CALLBACK (cb_sync), (gpointer) lfo_sync_check);
     gtk_widget_show (lfo_sync_check);

     /* sync menu */
     menu = gtk_menu_new ( );

     /* the different note size, the values are in beats */
     item = gtk_menu_item_new_with_label ("0.5 beats");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_sync_beats), (gpointer) EIGHTH);
     gtk_widget_show (item);

     item = gtk_menu_item_new_with_label ("1 beat");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_sync_beats), (gpointer) QUARTER);
     gtk_widget_show (item);

     item = gtk_menu_item_new_with_label ("2 beats");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_sync_beats), (gpointer) HALF);
     gtk_widget_show (item);

     item = gtk_menu_item_new_with_label ("4 beats");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_sync_beats), (gpointer) WHOLE);
     gtk_widget_show (item);

     item = gtk_menu_item_new_with_label ("8 beats");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_sync_beats), (gpointer) DOUBLE);
     gtk_widget_show (item);

     /* sync optionmenu */
     lfo_sync_opt = gtk_option_menu_new ( );
     gtk_box_pack_end (GTK_BOX (hbox), lfo_sync_opt, FALSE, FALSE, 0);
     gtk_option_menu_set_menu (GTK_OPTION_MENU (lfo_sync_opt), menu);
     gtk_widget_show (lfo_sync_opt);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /*********/
     /* Shape */
     /*********/
     label = gtk_label_new ("Shape:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* shape menu */
     menu = gtk_menu_new ( );

     /* triangle */
     item = gtk_menu_item_new_with_label ("Triangle");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (cb_shape),
		       (gpointer) LFO_SHAPE_TRIANGLE);
     gtk_widget_show (item);

     /* sine */
     item = gtk_menu_item_new_with_label ("Sine");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (cb_shape),
		       (gpointer) LFO_SHAPE_SINE);
     gtk_widget_show (item);

     /* saw */
     item = gtk_menu_item_new_with_label ("Saw");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (cb_shape),
		       (gpointer) LFO_SHAPE_SAW);
     gtk_widget_show (item);

     /* square */
     item = gtk_menu_item_new_with_label ("Square");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (cb_shape),
		       (gpointer) LFO_SHAPE_SQUARE);
     gtk_widget_show (item);

     /* shape optionmenu */
     lfo_shape_opt = gtk_option_menu_new ( );
     gtk_box_pack_end (GTK_BOX (hbox), lfo_shape_opt, FALSE, FALSE, 0);
     gtk_option_menu_set_menu (GTK_OPTION_MENU (lfo_shape_opt), menu);
     gtk_widget_show (lfo_shape_opt);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /**************/
     /* lfo amount */
     /**************/
     label = gtk_label_new ("Amount:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* adjustment */
     lfo_amt_adj = (GtkAdjustment *) gtk_adjustment_new (0.0, -1.0, 1.0, 0.01, 0.1, 0.0);
     lfo_pitch_amt_adj = (GtkAdjustment *) gtk_adjustment_new (0.0, -PATCH_MAX_PITCH_STEPS, PATCH_MAX_PITCH_STEPS,
							       0.01, 0.1, 0.0);
     g_signal_connect (G_OBJECT (lfo_amt_adj), "value-changed",
		       G_CALLBACK (cb_set), (gpointer) LFO_AMOUNT);
     g_signal_connect (G_OBJECT (lfo_pitch_amt_adj), "value-changed",
		       G_CALLBACK (cb_set), (gpointer) LFO_AMOUNT);
     /* spin */
     lfo_amt_spin = gtk_spin_button_new (lfo_amt_adj, 0, 2);
     gtk_box_pack_end (GTK_BOX (hbox), lfo_amt_spin, FALSE, FALSE, 0);
     gtk_widget_show (lfo_amt_spin);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /**********/
     /* Global */
     /**********/
     lfo_global_check = gtk_check_button_new_with_label ("Global" );
     gtk_box_pack_start (GTK_BOX (hbox), lfo_global_check, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (lfo_global_check), "toggled",
		       G_CALLBACK (cb_global), (gpointer) lfo_global_check);
     gtk_widget_show (lfo_global_check);

     /************/
     /* Positive */
     /************/
     lfo_positive_check = gtk_check_button_new_with_label ("Positive" );
     gtk_box_pack_end (GTK_BOX (hbox), lfo_positive_check, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (lfo_positive_check), "toggled",
		       G_CALLBACK (cb_positive), (gpointer) lfo_positive_check);
     gtk_widget_show (lfo_positive_check);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /***************** VEL PARAMS ********************/
     label = gtk_label_new (NULL);
     gtk_label_set_markup (GTK_LABEL (label), "<b>Velocity</b>");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /**************/
     /* vel amount */
     /**************/
     label = gtk_label_new ("Amount:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* adjustment */
     vel_amt_adj = (GtkAdjustment*) gtk_adjustment_new (0.0, 0.0, 1.0, 0.01, 0.1, 0.0);
     g_signal_connect (G_OBJECT (vel_amt_adj), "value-changed",
		       G_CALLBACK (cb_set), (gpointer) VEL_AMOUNT);

     /* spin */
     vel_amt_spin = gtk_spin_button_new (vel_amt_adj, 0, 2);
     gtk_box_pack_end (GTK_BOX (hbox), vel_amt_spin, FALSE, FALSE, 0);
     gtk_widget_show (vel_amt_spin);

     /**************/
     sep = gtk_hseparator_new ( );
     gtk_box_pack_start (GTK_BOX (vbox), sep, FALSE, FALSE, 0);
     gtk_widget_show (sep);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /* close button */
     button = gtk_button_new_from_stock (GTK_STOCK_CLOSE);
     g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (cb_close),
		       NULL);
     gtk_box_pack_end (GTK_BOX (hbox), button, FALSE, FALSE, 0);
     gtk_widget_show (button);

     /* TODO: this is kludgey, is there a better way? */
     g_object_ref (G_OBJECT (lfo_amt_adj));
     g_object_ref (G_OBJECT (lfo_pitch_amt_adj));
     g_object_ref (G_OBJECT (amt_adj));
     g_object_ref (G_OBJECT (pitch_amt_adj));
}
